package se.gawell.kvtstore.core.mysql;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;

import se.gawell.kvtstore.core.Deserializer;
import se.gawell.kvtstore.core.EntityNotFoundException;
import se.gawell.kvtstore.core.KeyTimeValueStore;
import se.gawell.kvtstore.core.SecondaryIndexValue;
import se.gawell.kvtstore.core.Serializer;
import se.gawell.kvtstore.core.TimeFormatter;
import se.gawell.kvtstore.core.TimeProvider;

public class MySQLKeyTimeValueStore implements KeyTimeValueStore<String> {

	Logger log = LoggerFactory.getLogger(getClass());
	
	public static final String CREATION_TIME_INDEX = "creation_time_index_bin";
	private final Serializer serializer;
	private final Deserializer deserializer;
	private final TimeProvider timeProvider;
	private final TimeFormatter timeFormatter;
	
	@Value("${jdbc.url:jdbc:mysql://localhost:3306/kvtstore}")
	private String jdbcUrl = "jdbc:mysql://localhost:3306/kvtstore";
	@Value("${jdbc.user:kvtstore}")
	private String jdbcUser = "kvtstore";
	@Value("${jdbc.password:kvtstore}")
	private String jdbcPassword = "kvtstore";

	@Autowired
	public MySQLKeyTimeValueStore(Serializer serializer,
			Deserializer deserializer, TimeProvider timeProvider,
			TimeFormatter timeFormatter) {
		super();
		this.serializer = serializer;
		this.deserializer = deserializer;
		this.timeProvider = timeProvider;
		this.timeFormatter = timeFormatter;
	}
	
	public void logConfiguration() {
		log.info("{} configuration: jdbc.url={} jbbc.user={} jdbc.password={}", getClass().getSimpleName(), jdbcUrl, jdbcUser, jdbcPassword);
	}

	@Override
	public String put(Object object, SecondaryIndexValue... indexes) {
		String primaryKey = UUID.randomUUID().toString();
		put(primaryKey, object, indexes);
		return primaryKey;
	}

	@Override
	public void put(String primaryKey, Object object,
			SecondaryIndexValue... indexes) {
		String serializedObject = serializer.serialize(object);
		boolean notExist = true;
		try {
			getById(primaryKey, object.getClass());
		} catch (EntityNotFoundException e) {
			notExist = false;
		}
		if (notExist) {
			executeStatement("UPDATE keyvalue SET data = '" + serializedObject
					+ "' WHERE primaryKey = '" + primaryKey + "' AND type='"
					+ object.getClass().getCanonicalName() + "';");
		} else {
			executeStatement("INSERT INTO keyvalue (primaryKey, type, data) VALUES('"
					+ primaryKey
					+ "','"
					+ object.getClass().getCanonicalName()
					+ "','" + serializedObject + "');");
		}

		executeStatement("DELETE FROM indexvalue WHERE type='"
				+ object.getClass().getCanonicalName() + "' AND primaryKey='"
				+ primaryKey + "';");

		putInIndex(primaryKey, object, CREATION_TIME_INDEX,
				timeFormatter.format(timeProvider.now()));
		for (SecondaryIndexValue secondaryIndexValue : indexes) {
			putInIndex(primaryKey, object, secondaryIndexValue.name,
					secondaryIndexValue.value);
		}
	}

	private void putInIndex(String primaryKey, Object object, String indexName,
			String indexValue) {
		executeStatement("INSERT INTO indexvalue (indexName, type, value, primaryKey) VALUES("
				+ "'"
				+ indexName
				+ "', "
				+ "'"
				+ getType(object)
				+ "', "
				+ "'"
				+ indexValue + "', " + "'" + primaryKey + "');");
	}

	private String getType(Object object) {
		return object.getClass().getCanonicalName();
	}

	@Override
	public <type> type getById(String primaryKey, Class<? extends Object> type) {
		try {
			Connection connection = getConnection();
			try {
				PreparedStatement preparedStatement = connection
						.prepareStatement("SELECT data from keyvalue WHERE primaryKey='"
								+ primaryKey
								+ "' and type='"
								+ type.getCanonicalName() + "';");
				ResultSet result = preparedStatement.executeQuery();
				result.first();
				String objectAsString = result.getString("data");
				return deserializer.deserialize(objectAsString, type);
			} finally {
				connection.close();
			}
		} catch (SQLException e) {
			throw new EntityNotFoundException(e.getMessage());
		}
	}

	@Override
	public <type> List<type> getByTimeInterval(Date startOfIntervalInclusive,
			Date endOfIntervalExclusive, Class<? extends Object> type) {
		String indexName = CREATION_TIME_INDEX;
		String startOfIntervalInclusiveAsString = timeFormatter
				.format(startOfIntervalInclusive);
		String endOfIntervalExclusiveAsString = timeFormatter
				.format(endOfIntervalExclusive);
		return getByIndexInterval(type, indexName,
				startOfIntervalInclusiveAsString,
				endOfIntervalExclusiveAsString);
	}

	private <type> List<type> getByIndexInterval(Class<? extends Object> type,
			String indexName, String startOfIntervalInclusiveAsString,
			String endOfIntervalExclusiveAsString) {
		try {
			Connection connection = getConnection();
			try {
				String genericSelectIndex = "SELECT primaryKey FROM indexvalue WHERE type='"
						+ type.getCanonicalName()
						+ "' AND indexName='"
						+ indexName + "'";
				String queryString = genericSelectIndex + " AND value >='"
						+ startOfIntervalInclusiveAsString + "';";
				if (endOfIntervalExclusiveAsString != null) {
					queryString = genericSelectIndex + " AND value >='"
							+ startOfIntervalInclusiveAsString
							+ "' AND value < '"
							+ endOfIntervalExclusiveAsString + "';";
				}
				logQuery(queryString);
				PreparedStatement preparedStatement = connection
						.prepareStatement(queryString);
				ResultSet result = preparedStatement.executeQuery();
				List<type> listOfEntities = new LinkedList<type>();
				while (result.next()) {
					String primaryKey = result.getString("primaryKey");
					listOfEntities.add((type) getById(primaryKey, type));
				}
				return listOfEntities;
			} finally {
				connection.close();
			}
		} catch (SQLException e) {
			throw new EntityNotFoundException(e.getMessage());
		}
	}


	@Override
	public <type> void delete(String primaryKey, Class<? extends Object> type) {
		executeStatement("DELETE FROM keyvalue WHERE primaryKey='" + primaryKey
				+ "' and type='" + type.getCanonicalName() + "';");
	}

	@Override
	public <type> List<type> getByIndex(SecondaryIndexValue index,
			Class<? extends Object> type) {
		return getByIndexInterval(type, index.name, index.value, null);
	}

	private void executeStatement(String statement) {
		logQuery(statement);
		try {
			Connection connection = getConnection();
			try {
				PreparedStatement preparedStatement = connection
						.prepareStatement(statement);
				preparedStatement.execute();
			} finally {
				connection.close();
			}
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}


	private Connection getConnection() throws SQLException {
		Connection connection = DriverManager.getConnection(
				jdbcUrl , jdbcUser , jdbcPassword );
		return connection;
	}

	private void logQuery(String queryString) {
		log.debug(queryString);
	}

}
